<?php

namespace FutureHome\Handler;

use FutureHome\Socket\Client;
use FutureHome\Socket\Logger;
use FutureHome\Socket\Message\SevenOfNine\Message;
use Ratchet\WebSocket\Version\RFC6455\Frame;
use React\Dns\Query\TimeoutException;
use React\Dns\Resolver\Factory as DnsFactory;

class Hyperchannel {
    protected $message_cache = array();
    protected $loop;

    protected $cube;
    protected $hive;
    protected $vinculum;
    protected $hive_connected = false;

    protected $versionablePackages = array(
            'futurehome',
            'futurehome-core',
            'futurehome-security',
            'nanoprobe',
            'unicomplex',
            'vinculum',
            'zipgateway',
            'zwave-ad',
            'fhtunnel',
            'fimpui',
        );

    protected $dns_servers;

    protected $versions = array();

    public function __construct($loop) {
        $this->loop = $loop;

        $this->update = new Update($this->loop);
        $this->update->on('download', array($this, 'handleDownloadMessage'));
        $this->update->on('updating', array($this, 'handleUpdatingMessage'));
        $this->update->on('update', array($this, 'handleUpdateMessage'));

        $this->log = new Log($this->loop);
    }

    public function initVinculum() {
        $this->vinculum = new Client\ClientVinculum($this->loop);
        $this->vinculum->on('connect', array($this, 'onVinculumConnected'));
        $this->vinculum->on('data', array($this, 'onVinculumData'));
        $this->vinculum->on('close', array($this, 'onVinculumDisconnect'));
        $this->vinculum->on('error', function ($ex) {
            $this->vinculum->disconnect();
            Logger::error('Vinculum, connection Error. Reason: '.$ex->getMessage());
            $this->loop->addTimer(5, [$this, 'initVinculum']);
        });
        $this->vinculum->connect();
    }

    public function initHive() {
        if (count($this->dns_servers) == 0) {
            Logger::error('Out of DNS servers to retry, exiting.');
            $this->loop->stop();
        } else {
            if ($this->hive) {
                $this->hive->removeAllListeners();
            }
            Logger::info("Connecting with DNS server: {$this->dns_servers[0]}");
            $factory  = new DnsFactory();
            $resolver = $factory->createCached($this->dns_servers[0], $this->loop);
            $this->hive = new Client\ClientHive($this->loop, $resolver);
            $this->hive->setVersions($this->versions);
            $this->hive->on('connected', array($this, 'onHiveConnected'));
            $this->hive->on('data', array($this, 'onHiveData'));
            $this->hive->on('close', array($this, 'onHiveDisconnect'));
            $this->hive->on('error', function ($ex) {
                try {
                    throw $ex;
                } catch (\Exception $ex) {
                    Logger::error('Hive Connection Error. Reason: '.$ex->getMessage() . ". Retrying in 2 seconds.");
                    array_shift($this->dns_servers);
                    $this->loop->addTimer(2, array($this, 'initHive'));
                }
            });
            $this->hive->connect();
        }
    }

    protected function getVersion($package) {
        $output = '';
        exec("dpkg -l $package | grep $package", $output);

        if (count($output) > 0) {
            $matches = array();
            $pattern = '/^(.*?) +(.*?) +(.*?) +(.*?) +(.*?)$/m';

            if (preg_match($pattern, $output[0], $matches) > 0) {
                Logger::info("Version for $package is ".$matches[3]);

                return $matches[3];
            }
        }

        return;
    }

    public function getVersions() {
        foreach ($this->versionablePackages as $package) {
            $this->versions[$package] = $this->getVersion($package);
        }
    }

    public function loadDNSFromResolveConf() {
        $resolve_conf_content = file_get_contents('/etc/resolv.conf');
        $dns = [];

        foreach(explode(PHP_EOL, $resolve_conf_content) as $line) {
            $line = trim($line);
            if (substr($line,0, 1) == "#") continue;

            $resolve_entry = explode(" ", $line);
            if (trim($resolve_entry[0]) == "nameserver") {
                $dns[] = trim($resolve_entry[1]);
            }
        }

        return $dns;
    }

    public function loadDNS() {
        $dns_from_config = \Config::get('app.dns_servers', []);

        $this->dns_servers = array_merge($this->loadDNSFromResolveConf(), $dns_from_config);
    }

    public function start() {
        $this->getVersions();
        $this->loadDNS();
        Logger::info("Using following DNS servers in specific order: " . implode(", ", $this->dns_servers));
        $this->initVinculum();
        $this->initHive();
    }

    public function onVinculumConnected() {
        if (isset($this->hive) && $this->hive) {
            $message = new Message((object) array());

            $message->setSrc(Message::SRC_HYPERCHANNEL);
            $message->setDst(Message::DST_CLIENT);
            $message->setType(Message::TYPE_NOTIFY);

            $message->setCommand(Message::COMMAND_SET);
            $message->setComponent(Message::COMPONENT_CUBE);

            $message->setParams(
                (object) array(
                    'message' => 'vinculum-online',
                )
            );

            $this->hive->send($message);
        }
    }

    public function onVinculumData($from, $msg) {
        if($this->hive && $this->hive->isConnected()) {
            $this->hive->send($msg);
        }
        else {
            Logger::info("Hive offline");
            if(!in_array($msg, $this->message_cache)) {
                $this->message_cache[] = $msg;
            }
        }
        if($this->cube) {
            $this->cube->send($msg);
        }
    }

    public function onVinculumDisconnect() {
        Logger::error('Vinculum got disconnected.');

        if ($this->hive && $this->hive->isConnected()) {
            $message = new Message((object) array());

            $message->setSrc(Message::SRC_HYPERCHANNEL);
            $message->setDst(Message::DST_CLIENT);
            $message->setType(Message::TYPE_NOTIFY);

            $message->setCommand(Message::COMMAND_SET);
            $message->setComponent(Message::COMPONENT_CUBE);

            $message->setParams(
                (object) array(
                    'message' => 'vinculum-offline',
                )
            );

            $this->hive->send($message);
        }

        $this->vinculum = null;
        $this->loop->addTimer(5, array($this, 'initVinculum'));
    }

    public function onHiveData($from, $msg) {
        $message = new Message(json_decode($msg, false));

        if ($message->getDst() == Message::DST_HYPERCHANNEL) {
            if ($message->getType() == Message::TYPE_REQUEST) {
                if ($message->getComponent() == Message::COMPONENT_UPDATE) {
                    if ($message->getCommand() == Message::COMMAND_EDIT) {
                        // Later this might be used to instruct specific version to be downloaded
                        $this->update->runUpdateDownloader();
                        $m = new Message((object) array());
                        $m->setSrc(Message::SRC_HYPERCHANNEL);
                        $m->setDst($message->getSrc());
                        $m->setType(Message::TYPE_RESPONSE);
                        $m->setComponent($message->getComponent());
                        $m->setCommand($message->getCommand());
                        $m->setParams((object) array());
                        $m->setRequestId($message->getRequestId());
                        $m->setSuccess(true);
                        if (isset($this->hive) && $this->hive) {
                            $this->hive->send($m);
                        }
                    } else if ($message->getCommand() == Message::COMMAND_SET) {
                        $this->update->runUpdate();
                        $m = new Message((object) array());
                        $m->setSrc(Message::SRC_HYPERCHANNEL);
                        $m->setDst($message->getSrc());
                        $m->setType(Message::TYPE_RESPONSE);
                        $m->setComponent($message->getComponent());
                        $m->setCommand($message->getCommand());
                        $m->setParams((object) array());
                        $m->setRequestId($message->getRequestId());
                        $m->setSuccess(true);
                        if (isset($this->hive) && $this->hive) {
                            $this->hive->send($m);
                        }
                    }
                } else if ($message->getComponent() == Message::COMPONENT_DEBUG) {
                    if ($message->getCommand() == Message::COMMAND_GET) {
                        $param = $message->getParams();
                        if (
                            !property_exists($param, 'type') ||
                            !in_array($param->type, array('hyperchannel', 'vinculum'))
                        ) {
                            $m = new Message((object) array());
                            $m->setSrc(Message::SRC_HYPERCHANNEL);
                            $m->setDst($message->getSrc());
                            $m->setType(Message::TYPE_RESPONSE);
                            $m->setComponent($message->getComponent());
                            $m->setCommand($message->getCommand());
                            $m->setParams((object) array());
                            $m->setRequestId($message->getRequestId());
                            $m->setSuccess(false);
                            $m->setError(array(
                                'msg' => 'Not a valid type'
                            ));
                            if (isset($this->hive) && $this->hive) {
                                $this->hive->send($m);
                            }
                        } else {
                            $log = false;
                            if ($param->type == 'hyperchannel') {
                                $log = $this->log->getHyperchannelLog();
                            } else if ($param->type == 'vinculum') {
                                $log = $this->log->getVinculumLog();
                            }
                            if ($log != false) {
                                $m = new Message((object) array());
                                $m->setSrc(Message::SRC_HYPERCHANNEL);
                                $m->setDst($message->getSrc());
                                $m->setType(Message::TYPE_RESPONSE);
                                $m->setComponent($message->getComponent());
                                $m->setCommand($message->getCommand());
                                $m->setParams((object)$log);
                                $m->setRequestId($message->getRequestId());
                                $m->setSuccess(true);
                                if (isset($this->hive) && $this->hive) {
                                    $this->hive->send($m);
                                }
                            } else {
                                Logger::error("Log request failed.");
                                $m = new Message((object) array());
                                $m->setSrc(Message::SRC_HYPERCHANNEL);
                                $m->setDst($message->getSrc());
                                $m->setType(Message::TYPE_RESPONSE);
                                $m->setComponent($message->getComponent());
                                $m->setCommand($message->getCommand());
                                $m->setParams((object) array());
                                $m->setRequestId($message->getRequestId());
                                $m->setSuccess(false);
                                if (isset($this->hive) && $this->hive) {
                                    $this->hive->send($m);
                                }
                            }
                        }
                    }
                }
            }

            return;
        }
        if ($this->vinculum && $this->vinculum->isConnected()) {
            $this->vinculum->send($msg);
        } else {
            if ($this->hive && $this->hive->isConnected()) {
                $message = new Message((object) array());

                $message->setSrc(Message::SRC_HYPERCHANNEL);
                $message->setDst(Message::DST_CLIENT);
                $message->setType(Message::TYPE_NOTIFY);

                $message->setCommand(Message::COMMAND_SET);
                $message->setComponent(Message::COMPONENT_CUBE);

                $message->setParams(
                    (object) array(
                        'message' => 'vinculum-offline',
                    )
                );

                $this->hive->send($message);
            }
        }
    }

    public function onHiveConnected() {
        if (count($this->message_cache) > 0) {
            foreach ($this->message_cache as $message) {
                $this->hive->send($message);
            }
        }
        $this->hive_connected = true;
    }

    public function onHiveDisconnect() {
        Logger::error('Hive got disconnected. Retrying in 2 seconds');

        try {
            throw new \Exception("Throwing exception to figure out disconnection path");
        } catch (\Exception $ex) {
            Logger::error('Hive Connection Error. Reason: '.$ex->getMessage());
            Logger::error('Trace: '.$ex->getTraceAsString());
        }
        // Restart hyperchannel if hive disconnects
        //$this->hive = null;
        if ($this->hive_connected) {
            Logger::error('Hive was already connected once, restarting it to not get stuck in some sort weird of loop');
            $this->loop->stop();
        }
        $this->loop->addTimer(2, array($this, 'initHive'));
    }

    public function handleUpdatingMessage($event) {
        $message = new Message((object) array());
        $message->setSrc(Message::SRC_UPDATE);
        $message->setDst(Message::DST_CLIENT);
        $message->setType(Message::TYPE_NOTIFY);
        $message->setComponent(Message::COMPONENT_UPDATE);
        $message->setCommand(Message::COMMAND_UPDATE);
        $updates = array(
            'update' => $event,
        );
        $message->setParams((object) $updates);
        if ($this->hive && $this->hive->isConnected()) {
            $this->hive->send($message);
        }
    }

    public function handleDownloadMessage($event) {
        $message = new Message((object) array());
        $message->setSrc(Message::SRC_UPDATE);
        $message->setDst(Message::DST_CLIENT);
        $message->setType(Message::TYPE_NOTIFY);
        $message->setComponent(Message::COMPONENT_UPDATE);
        $message->setCommand(Message::COMMAND_UPDATE);
        $updates = array(
            'update' => $event,
        );
        $message->setParams((object) $updates);
        if ($this->hive && $this->hive->isConnected()) {
            $this->hive->send($message);
        }
    }

    public function handleUpdateMessage($updates, $escapepodupdate) {
        $total_update = $updates + $escapepodupdate;
        $message = new Message((object) array());
        $message->setSrc(Message::SRC_UPDATE);
        $message->setDst(Message::DST_CLIENT);
        $message->setType(Message::TYPE_NOTIFY);
        $message->setComponent(Message::COMPONENT_UPDATE);
        $message->setCommand(Message::COMMAND_UPDATE);
        $updates = array(
                'count' => $total_update,
            );
        $message->setParams((object) $updates);
        if ($this->hive && $this->hive->isConnected()) {
            $this->hive->send($message);
        }
    }
}
