<?php

namespace FutureHome\Socket\Client;

use FutureHome\Socket\Logger;
use FutureHome\Socket\Message\SevenOfNine\Message;
use FutureHome\WSClient\Connector;
use FutureHome\WSClient\WebSocket;
use React\EventLoop\LoopInterface;
use React\Dns\Resolver\Resolver;
use React\HttpClient\Client as HttpClient;
use Evenement\EventEmitterInterface;
use Evenement\EventEmitterTrait;
use Ratchet\RFC6455\Messaging\MessageInterface;
use React\Dns\Resolver\Factory as DnsFactory;
use React\Socket\Connector as ReactConnector;

class ClientHive extends Connector implements EventEmitterInterface {
    use EventEmitterTrait;
    protected $versions;

    protected $timeout_timer;

    protected $connected = false;

    protected $connection;

    protected $resolver;

    public function __construct(LoopInterface $loop, Resolver $resolver = null, array $secureContext = []) {
        parent::__construct($loop, $resolver, $secureContext);
        $this->resolver = $resolver;
    }

    public function isConnected() {
        return $this->connected;
    }

    public function disconnect() {
        if ($this->connected === true) {
          if ($this->connection) {
              $this->connection->close();
          }
          $this->connected = false;
        }
        $this->emit('close');
    }

    public function setVersions($versions) {
        $this->versions = $versions;
    }

    public function connect() {
        if ($this->timeout_timer) {
            $this->loop->cancelTimer($this->timeout_timer);
        }
        $this->timeout_timer = $this->loop->addTimer(90, function() {
            Logger::line("Disconnecting hive connection because session could not be generated.");
            $this->disconnect();
        });
        $this->on('registered', function($registration_data) {
            Logger::info("Registered to hive Successfully. Connecting after " . $registration_data['connect_after']);
            if ($this->timeout_timer) {
                $this->loop->cancelTimer($this->timeout_timer);
            }
            $this->loop->addTimer($registration_data['connect_after'], function() use($registration_data) {
                if ($this->timeout_timer) {
                    $this->loop->cancelTimer($this->timeout_timer);
                }
                $this->timeout_timer = $this->loop->addTimer(120, function() {
                    Logger::line("Disconnecting hive connection because a stable websocket connection could not be established.");
                    $this->disconnect();
                });
                $this(sprintf("%s://%s:%s%s", $registration_data['protocol'], $registration_data['hostname'], $registration_data['port'],$registration_data['path']), [], [], $this->loop)->then(function(WebSocket $conn) {
                    Logger::info('Connected to server.');
                    if ($this->timeout_timer) {
                        $this->loop->cancelTimer($this->timeout_timer);
                    }
                    $this->timeout_timer = $this->loop->addTimer(90, function() {
                        if ($this->connected) {
                            Logger::line("Disconnecting hive connection, first ping not received.");
                            $this->disconnect();
                        }
                    });
                    $this->connected = true;
                    $this->connection = $conn;
                    $conn->on('message', [$this, 'onMessage']);
                    $conn->on('close', function() {
                        $this->connection = null;
                        $this->disconnect();
                    });
                    $conn->on('error', function($error) {
                        $this->disconnect();
                    });
                    $this->initSite();
                    $this->emit('connected');
                }, function(\Exception $e) {
                    $this->emit('error', [$e]);
                    Logger::error($e->getMessage());
                    Logger::error($e->getTraceAsString());
                    $this->emit('close');
                });
            });
        });
        $this->createSession();
    }

    public function send($message) {
        return $this->connection->send($message);
    }

    protected function createSession() {
        $registration_uri = \Config::get('app.hive').'v1/cube/register';
        $connector = new ReactConnector(
            $this->loop,
            [
                'dns' => $this->resolver
            ]
        );
        $client = new HttpClient($this->loop, $connector);

        $request = $client->request('GET', $registration_uri);
        $registration_data = '';
        $request->on('response', function ($response) use(&$registration_data) {
            $response->on('data', function ($chunk) use(&$registration_data) {
                $registration_data .= $chunk;
            });
            $response->on('end', function() use(&$registration_data) {
                $registration = json_decode($registration_data, true);
                if (!is_array($registration)) {
                    Logger::error("Couldn't register to the hive.");

                    $this->emit('error', array(new \Exception("Could not register to hive.")));
                }
                if (!array_key_exists('protocol', $registration)) {
                    $registration['protocol'] = 'wss';
                    if (!array_key_exists('hostname', $registration)) {
                        Logger::error("Hive skipped the registration, hyperchannel restart is required.");
                        exit();
                    }
                    $registration['hostname'] = preg_replace('#^ssl://#', '', $registration['hostname']);
                } else if ($registration['protocol'] == 'ssl') {
                    $registration['protocol'] = 'wss';
                }
                if (array_key_exists('path', $registration)) {
                    $registration['path'] = '/'.$registration['path'];
                } else {
                    $registration['path'] = '/vinculum';
                }
                if (!array_key_exists('connect_after', $registration)) {
                    $registration['connect_after'] = rand(10, 20);
                    Logger::info(sprintf("Setting connect after value to %d", $registration['connect_after']));
                } else {
                    Logger::info(sprintf("Got connect after value as %d", $registration['connect_after']));
                }
                $this->emit('registered', array($registration));
            });
        });

        $request->on('error', function (\Exception $e) {
            Logger::error($e->getMessage());
            Logger::error($e->getTraceAsString());
            $this->emit('error', array($e));
        });
        $request->end();
    }

    protected function deleteSite() {
        $filesystem = new \Illuminate\Filesystem\Filesystem();

        $filesystem->delete(\Config::get('app.data_folder_path').'site.data');
    }

    protected function setupSite($siteId) {
        $filesystem = new \Illuminate\Filesystem\Filesystem();

        $filesystem->put(\Config::get('app.data_folder_path').'site.data', $siteId);
    }

    protected function initSite() {
        $filesystem = new \Illuminate\Filesystem\Filesystem();

        try {
            $siteId = $filesystem->get(\Config::get('app.data_folder_path').'site.data');

            $message = new Message((object) array(
                'data' => (object) array(
                    'cmd' => Message::COMMAND_SET,
                    'component' => Message::COMPONENT_CUBE,
                    'requestId' => rand(),
                    'param' => (object) array(
                        'siteId' => trim($siteId),
                        'version' => $this->versions,
                    ),
                ),
            ));

            $message->setSrc(Message::SRC_HYPERCHANNEL);
            $message->setDst(Message::DST_HIVE);

            $message->setType(Message::TYPE_REQUEST);

            $this->connection->send($message);
        } catch (\Exception $ex) {
            Logger::error($ex->getMessage());
        }
    }

    public function onMessage($msg, $from) {
        try {
            $msg = $msg->getPayload();

            $m = json_decode($msg, false);
            if (!is_object($m)) {
                Logger::error('Received invalid message'.$msg);

                return;
            }
            $message = new Message($m);

            if ($message->getDst() == Message::DST_HYPERCHANNEL) {
                // These are the messages to be handled by Hyperchannel itself
                    // Don't pass them anywhere
                    Logger::comment('Got Hyperchannel message: '.$message);
                if ($message->getType() == Message::TYPE_REQUEST) {
                    // These type of messages always need to have a response
                    if ($message->getComponent() == Message::COMPONENT_CUBE) {
                        if ($message->getCommand() == Message::COMMAND_LRN) {
                            $data = $message->getData();

                            $siteId = $data->param->identifier;
                            $requestId = $data->requestId;

                            $this->setupSite($siteId);

                            $this->initSite();

                            $dst = $message->getSrc();

                            $message->setType(Message::TYPE_NOTIFY);
                            $message->setSrc($message->getDst());
                            $message->setDst(Message::DST_CLIENT);

                            $data = (object) array(
                                'component' => Message::COMPONENT_CUBE,
                                'cmd' => Message::COMMAND_LRN,
                                'id' => $siteId,
                                'requestId' => $requestId,
                                'success' => true,
                                'param' => (object) array(
                                    'learned' => true,
                                ),
                            );

                            $message->setData((object) $data);

                            // Notify all the users that a new cube was added
                            $this->send($message);

                            $message->setType(Message::TYPE_RESPONSE);
                            $message->setDst($dst);

                            $this->connection->send($message);
                        }
                        return;
                    }
                } else if ($message->getType() == Message::TYPE_PING) {
                    if ($this->timeout_timer) {
                        $this->loop->cancelTimer($this->timeout_timer);
                    }
                    $this->timeout_timer = $this->loop->addTimer(90, function() {
                        if ($this->connected) {
                            Logger::line("Disconnecting hive connection because no message was received for 90 seconds.");
                            $this->disconnect();
                        }
                    });
                    Logger::line("Got Ping from Hive");
                    return;
                }
            }
            if ($message->getType() == Message::TYPE_REQUEST) {
                if ($message->hasComponent(Message::COMPONENT_CONFIG)) {
                    if ($message->getCommand() == Message::COMMAND_DELETE) {
                        $this->deleteSite();
                        $this->loop->addTimer(5, function() {
                            $this->connection->close();
                        });
                    }
                }
            }
            $this->emit('data', array($from, $msg));
        } catch (\Exception $ex) {

        }
    }
}
