<?php
namespace FutureHome\WSClient;

use Evenement\EventEmitterTrait;
use Evenement\EventEmitterInterface;
use React\Stream\DuplexStreamInterface;
use Psr\Http\Message\RequestInterface;
use Psr\Http\Message\ResponseInterface;
use Ratchet\RFC6455\Messaging\MessageBuffer;
use Ratchet\RFC6455\Messaging\CloseFrameChecker;
use Ratchet\RFC6455\Messaging\MessageInterface;
use Ratchet\RFC6455\Messaging\FrameInterface;
use Ratchet\RFC6455\Messaging\Frame;

class WebSocket implements EventEmitterInterface {
    use EventEmitterTrait;
    /**
     * The request headers sent to establish the connection
     * @var \Psr\Http\Message\RequestInterface
     */
    public $request;
    /**
     * The response headers received from the server to establish the connection
     * @var \Psr\Http\Message\ResponseInterface
     */
    public $response;
    /**
     * @var \React\Stream\Stream
     */
    protected $stream;
    /**
     * @var \Closure
     */
    protected $close;

    /**
     * WebSocket constructor.
     * @param \React\Stream\DuplexStreamInterface $stream
     * @param \Psr\Http\Message\ResponseInterface $response
     * @param \Psr\Http\Message\RequestInterface  $request
     * @event message
     * @event pong
     * @event close
     * @event error
     */
    public function __construct(DuplexStreamInterface $stream, ResponseInterface $response, RequestInterface $request) {
        $this->stream  = $stream;

        /*
        * Hack to set writeChunkSize to some sane value
        */
        $stream_reflection = new \ReflectionClass(get_class($this->stream));
        $duplex_resource_stream_variable = $stream_reflection->getProperty('input');
        $duplex_resource_stream_variable->setAccessible(true);
        $duplex_resource_stream = $duplex_resource_stream_variable->getValue($this->stream);

        $duplex_resource_stream_reflection = new \ReflectionClass(get_class($duplex_resource_stream));
        $writable_resource_stream_variable = $duplex_resource_stream_reflection->getProperty('buffer');
        $writable_resource_stream_variable->setAccessible(true);
        $writable_resource_stream = $writable_resource_stream_variable->getValue($duplex_resource_stream);

        $writable_resource_stream_reflection = new \ReflectionClass(get_class($writable_resource_stream));
        $write_chunk_size_variable = $writable_resource_stream_reflection->getProperty('writeChunkSize');
        $write_chunk_size_variable->setAccessible(true);
        $write_chunk_size_variable->setValue($writable_resource_stream, 2048);
        /*
        * End hack to set writeChunkSize to some sane value
        */

        $this->response = $response;
        $this->request  = $request;
        $self = $this;
        $this->close = function($code = null, $reason = null) use ($self) {
            static $sent = false;
            if ($sent) {
                return;
            }
            $sent = true;
            $self->emit('close', [$code, $reason, $self]);
        };
        $reusableUAException = new \UnderflowException;
        $streamer = new MessageBuffer(
            new CloseFrameChecker,
            function(MessageInterface $msg) {
                $this->emit('message', [$msg, $this]);
            },
            function(FrameInterface $frame) use (&$streamer) {
                switch ($frame->getOpcode()) {
                    case Frame::OP_CLOSE:
                        $frameContents = $frame->getPayload();
                        $reason = '';
                        $code = unpack('n', substr($frameContents, 0, 2));
                        $code = reset($code);
                        if (($frameLen = strlen($frameContents)) > 2) {
                            $reason = substr($frameContents, 2, $frameLen);
                        }
                        $closeFn = $this->close;
                        $closeFn($code, $reason);
                        return $this->stream->end($streamer->newFrame($frame->getPayload(), true, Frame::OP_CLOSE)->maskPayload()->getContents());
                    case Frame::OP_PING:
                        return $this->send($streamer->newFrame($frame->getPayload(), true, Frame::OP_PONG));
                    case Frame::OP_PONG:
                        return $this->emit('pong', [$frame, $this]);
                    default:
                        return $this->close(Frame::CLOSE_PROTOCOL);
                }
            },
            false,
            function() use ($reusableUAException) {
                return $reusableUAException;
            }
        );
        $stream->on('data', [$streamer, 'onData']);
        $stream->on('end', function() use ($stream) {
            if (is_resource($stream->stream)) {
                stream_socket_shutdown($stream->stream, STREAM_SHUT_RDWR);
                stream_set_blocking($stream->stream, false);
            }
        });
        $stream->on('close', function () {
            $close = $this->close;
            $close(Frame::CLOSE_ABNORMAL, 'Underlying connection closed');
        });
        $stream->on('error', function($error) {
            $this->emit('error', [$error, $this]);
        });
    }

    public function send($msg) {
        if ($msg instanceof MessageInterface) {
            foreach ($msg as $frame) {
                $frame->maskPayload();
            }
        } else {
            if (!($msg instanceof Frame)) {
                $msg = new Frame($msg);
            }
            $msg->maskPayload();
        }
        $this->stream->write($msg->getContents());
    }

    public function close($code = 1000, $reason = '') {
        $frame = new Frame(pack('n', $code) . $reason, true, Frame::OP_CLOSE);
        $this->stream->write($frame->getContents());
        $closeFn = $this->close;
        $closeFn($code, $reason);
        $this->stream->end();
    }
}
